/**
 * @package koa-files
 * @license MIT
 * @version 2.3.8
 * @author nuintun <nuintun@qq.com>
 * @description A static files serving middleware for koa.
 * @see https://github.com/nuintun/koa-files#readme
 */

import fs from 'fs';
import { relative, isAbsolute } from 'path';

/**
 * @module utils
 * @license MIT
 * @author nuintun
 */
const CHARS = Array.from('0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz');
/**
 * @function isOutRoot
 * @description Check if path is out of root
 * @param path The path to check
 * @param root The root path
 */
function isOutRoot(path, root) {
  path = relative(root, path);
  return /\.\.(?:[\\/]|$)/.test(path) || isAbsolute(path);
}
/**
 * @function unixify
 * @description Convert path to unix style
 * @param path The path to convert
 */
function unixify(path) {
  return path.replace(/\\/g, '/');
}
/**
 * @function decodeURI
 * @description Decode URI component
 * @param URI The URI to decode
 */
function decodeURI(URI) {
  try {
    return decodeURIComponent(URI);
  } catch (error) {
    return -1;
  }
}
/**
 * @function boundaryGenerator
 * @description Generate a boundary
 */
function boundaryGenerator() {
  let boundary = '';
  // Create boundary
  for (let i = 0; i < 38; i++) {
    boundary += CHARS[Math.floor(Math.random() * 62)];
  }
  // Return boundary
  return boundary;
}
/**
 * @function fstat
 * @description Get file stats
 * @param path The file path
 */
function fstat(path) {
  return new Promise((resolve, reject) => {
    fs.stat(path, (error, stats) => {
      error ? reject(error) : resolve(stats);
    });
  });
}
/**
 * @function hasTrailingSlash
 * @description Check if path has trailing slash
 * @param path The path to check
 */
function hasTrailingSlash(path) {
  return /\/$/.test(path);
}
/**
 * @function parseTokens
 * @description Parse HTTP tokens
 * @param value The tokens value string
 */
function parseTokens(value) {
  let end = 0;
  let start = 0;
  let tokens = [];
  // gather tokens
  for (let i = 0, length = value.length; i < length; i++) {
    switch (value.charCodeAt(i)) {
      case 0x20:
        // ' '
        if (start === end) {
          start = end = i + 1;
        }
        break;
      case 0x2c:
        // ','
        tokens.push(value.substring(start, end));
        start = end = i + 1;
        break;
      default:
        end = i + 1;
        break;
    }
  }
  // final token
  tokens.push(value.substring(start, end));
  return tokens;
}
/**
 * @function isETag
 * @description Check if etag is valid
 * @param value The value to check
 */
function isETag(value) {
  return /^(?:W\/)?"[\s\S]+"$/.test(value);
}
/**
 * @function isETagFresh
 * @description Check if etag is fresh
 * @param match The match value
 * @param etag The etag value
 */
function isETagFresh(match, etag) {
  return parseTokens(match).some(match => {
    return match === etag || match === 'W/' + etag || 'W/' + match === etag;
  });
}

export { boundaryGenerator, decodeURI, fstat, hasTrailingSlash, isETag, isETagFresh, isOutRoot, parseTokens, unixify };
